using Microsoft.Xna.Framework;

namespace KavTest
{
    public struct Transform3D : System.IEquatable<Transform3D>
    {
        public Matrix TransformMatrix { get; private set; }

        public Vector3 Position { get; private set; }

        public Vector3 Forward { get { return TransformMatrix.Forward; } }

        public Transform3D(Vector3 position, Quaternion orientation, Vector3 scale)
        {
            Position = position;
            TransformMatrix = CreateTransformMatrix(position, orientation, scale);
        }

        public Transform3D(Vector3 position)
        {
            Position = position;
            TransformMatrix = CreateTransformMatrix(position, Quaternion.Identity, Vector3.One);
        }

        public Transform3D(Matrix matrix)
        {
            Position = matrix.Translation;
            TransformMatrix = matrix;
        }

        public Transform3D WithPosition(Vector3 position)
        {
            return Translate(position - TransformMatrix.Translation);
        }

        public Transform3D Translate(Vector3 translation)
        {
            return new Transform3D(Matrix.CreateTranslation(translation) * TransformMatrix);
        }

        public Transform3D Rotate(Quaternion other)
        {
            return new Transform3D(Matrix.CreateFromQuaternion(other) * TransformMatrix);
        }

        public Transform3D Rotate(float yaw, float pitch, float roll)
        {
            return Rotate(Quaternion.CreateFromYawPitchRoll(
                yaw,
                pitch,
                roll
            ));
        }

        public Transform3D Rotate(Vector3 eulerAngles)
        {
            return Rotate(Quaternion.CreateFromYawPitchRoll(
                eulerAngles.X,
                eulerAngles.Y,
                eulerAngles.Z
            ));
        }

        public Transform3D Compose(Transform3D other)
        {
            return new Transform3D(
                other.TransformMatrix * TransformMatrix
            );
        }

        private static Matrix CreateTransformMatrix(Vector3 position, Quaternion orientation, Vector3 scale)
        {
            return Matrix.CreateScale(scale) *
                   Matrix.CreateFromQuaternion(orientation) *
                   Matrix.CreateTranslation(position);
        }

        public override bool Equals(object other)
        {
            if (other is Transform3D otherTransform)
            {
                return Equals(otherTransform);
            }

            return false;
        }

        public bool Equals(Transform3D other)
        {
            return TransformMatrix == other.TransformMatrix;
        }

        public override int GetHashCode()
        {
            return TransformMatrix.GetHashCode();
        }

        public static bool operator ==(Transform3D a, Transform3D b)
        {
            return a.Equals(b);
        }

        public static bool operator !=(Transform3D a, Transform3D b)
        {
            return !(a == b);
        }
    }
}
